;;;; Perl-in.lisp - Lisp interface to the Perl API

;;; Copyright 2006 Stuart Sierra

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the Lisp Lesser GNU General Public
;;; License (LLGPL) published by Franz, Inc., available at
;;; http://opensource.franz.com/preamble.html

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; Lesser GNU General Public License for more details.

(cl:in-package :common-lisp-user)

(defpackage :perl-in-lisp
  (:use :common-lisp :cffi :cffi-address :perl-api)
  (:nicknames :perl)
  (:export #:start-perl #:stop-perl
           #:call-perl
           #:use-perl-module))

(in-package :perl-in-lisp)
(defvar *perl-interpreter* nil)
(defvar *sv-type-mask* #Xff)
(defmacro perl-scope (&body body)
  (let ((return-symbol (gensym)))
    `(progn 
       (perl-push-scope)  ; ENTER
       ;; SAVETMPS omitted
       (let ((,return-symbol (progn ,@body))) ; execute body
         (perl-free-tmps) ; FREETMPS
         (perl-pop-scope) ; LEAVE
         ,return-symbol))))
(defun make-interpreter ()
  (let ((interpreter (perl-alloc))
        (arguments (foreign-alloc :pointer :count 3)))
    (perl-construct interpreter)
    (setf (mem-aref arguments :pointer 0) (foreign-string-alloc ""))
    (setf (mem-aref arguments :pointer 1) (foreign-string-alloc "-e"))
    (setf (mem-aref arguments :pointer 2) (foreign-string-alloc "0"))
    (perl-parse interpreter (null-pointer)
                3 arguments (null-pointer))
    (let ((run (perl-run interpreter)))
      (unless (zerop run)
        (error "perl_run failed (return value: ~A)" run)))
    interpreter))
(defun destroy-interpreter (interpreter)
  (setf *pl-perl-destruct-level* 1)
  (perl-destruct interpreter)
  (perl-free interpreter))
(declaim (inline need-perl))

(defun need-perl ()
  (unless *perl-interpreter* (start-perl)))
(defun refcnt (scalar)
  (foreign-slot-value scalar 'sv 'refcnt))
(defun svtype (scalar)
  (foreign-enum-keyword
   'svtype
   (logand (foreign-slot-value scalar 'sv 'flags)
           *sv-type-mask*)))
(defun string-from-sv (sv)
  (with-foreign-object (length :strlen)
    (foreign-string-to-lisp
        (perl-sv-pvn-force-flags sv length
                                 (foreign-bitfield-value
                                  'sv-flags '(:gmagic)))
        (mem-ref length :strlen)
        nil))) ; null-teminated-p
(defun set-sv (destination source &rest flags)
  (perl-sv-setsv-flags destination source
                       (foreign-bitfield-value
                        'sv-flags flags)))
(defun get-scalar-by-name (name)
  (need-perl)
  (perl-get-sv name t))
(defun av-fetch-sv (array key create)
  (let ((ptr (perl-av-fetch array key create)))
    (if (null-pointer-p ptr) nil
      (mem-ref ptr :pointer))))
(defun perl-aref (array index)
  (need-perl)
  (lisp-from-perl (av-fetch-sv array index t)))
(defun list-from-av (av)
  (loop for i from 0 upto (perl-av-len av)
        collecting (perl-aref av i)))
(defun hash-from-hv (perl-hash)
  (perl-scope ;; because SVs may be mortal copies
   (let ((lisp-hash (make-hash-table :test #'equal))
         (size (perl-hv-iterinit perl-hash)))
     (loop repeat size
           do (let ((entry (perl-hv-iternext perl-hash)))
                (setf (gethash (string-from-sv ; does not work w/ lisp-from-perl, why?
                                (perl-hv-iterkeysv entry))
                               lisp-hash)
                      (lisp-from-perl
                       (perl-hv-iterval perl-hash entry)))))
     lisp-hash)))
(defun hv-from-hash (lisp-hash)
  (let ((perl-hash (perl-newhv)))
    (maphash #'(lambda (key value)
                 (let ((string-key (string key)))
                   (with-foreign-string
                    (cstring string-key)
                    (perl-hv-store perl-hash
                                   cstring
                                   (length string-key)
                                   (perl-from-lisp value)
                                   0))))
             lisp-hash)
    perl-hash))
(defun deref-rv (ref)
  (foreign-slot-value (foreign-slot-value ref 'sv 'any)
                      'xrv 'rv))
(defun pushmark ()
  (when (= (address-incf *pl-markstack-ptr*) *pl-markstack-max*)
    (perl-markstack-grow))
  (setf (address-ref *pl-markstack-ptr* :address)
        (- *pl-stack-sp* *pl-stack-base*)))
(defun ensure-room-on-stack (n)
  (when (< (- *pl-stack-max* *pl-stack-sp*) n)
    (setf *pl-stack-sp*
          (perl-stack-grow *pl-stack-sp* *pl-stack-sp* n))))
(defun pushs (scalar)
  "Push a scalar value (a pointer) onto the Perl stack."
  (ensure-room-on-stack 1) ; EXTEND
  (setf (address-ref (address-incf *pl-stack-sp*) :address)
        (pointer-address scalar)))
(defun pops ()
  "Pop a scalar pointer off the Perl stack."
  (prog1
      (address-ref *pl-stack-sp* :pointer)
    (address-decf *pl-stack-sp*)))
(defun push-mortals-on-stack (args)
  (loop for arg in args
        do (pushs (perl-sv-2mortal (perl-from-lisp arg)))))
(defun get-from-stack (n)
  (nreverse (loop repeat n
                  collecting (lisp-from-perl (pops)))))

(defun get-stack-values (n)
  (values-list (get-from-stack n)))
(defun perl-call-scalar (scalar flags)
  (perl-call-sv scalar (foreign-bitfield-value
                         'perl-call-flags flags)))

(defun perl-call-function (name flags)
  (perl-call-pv name (foreign-bitfield-value
                         'perl-call-flags flags)))
(defun context-from-type (type)
  (cond
   ((null type) :void)
   ((find type '(:integer :float :string t)) :scalar)
   ((find type '(:list :array :alist :hash)) :array)
   (t (error "No Perl calling context for type ~a" type))))
(defun calling-flags (type methodp)
  (let ((flags (list :eval (context-from-type type))))
    (when methodp (push :method flags))
    flags))
(defun get-stack-by-type (type count)
  (declare (ignore type)) ;; fix me
  (get-stack-values count))
(defun perl-eval-scalar (scalar flags)
  (perl-eval-sv scalar (foreign-bitfield-value
                        'perl-call-flags flags)))
(defun eval-perl (code)
  (need-perl)
  (perl-scope
   (get-stack-values
    (perl-eval-scalar (perl-sv-2mortal (perl-newsvpv code 0))
                      '(:scalar :eval)))))
(defgeneric perl-from-lisp (value))

(defmethod perl-from-lisp ((value integer))
  (need-perl)
  (cond
   ((and (<= 0 value 4294967295)) ;; 32-bit unsigned integers
    (perl-newsvuv value))
   ((and (> 0 value -2147483648)) ;; 32-bit signed integers
    (perl-newsviv value))
   (t (error "Integer value out of range for Perl;
BigInts not supported"))))

(defmethod perl-from-lisp ((value float))
  (need-perl)
  (perl-newsvnv
   ;; ensure VALUE is a double-float
   (float value 1.0d0)))

(defmethod perl-from-lisp ((value string))
  (need-perl)
  (perl-newsvpv value 0))

(defmethod perl-from-lisp ((value list))
  (let ((a (perl-newav)))
    (loop for i in value
          ;; Perl's "push" pushes to the *end* of the array
          do (perl-av-push a (perl-from-lisp i)))
    a))

(defmethod perl-from-lisp ((value hash-table))
  (hv-from-hash value))
(defmethod perl-from-lisp ((value null)) ; NIL isn't a class; NULL is
  (perl-newsv 0))
(defun lisp-from-perl (p)
  (ecase (svtype p)
    (:null nil)
    (:iv (perl-sv-2iv p))
    (:nv (perl-sv-2nv p))
    (:rv (lisp-from-perl (deref-rv p)))
    (:pv (string-from-sv p))
    (:pviv (cffi:mem-ref p :iv))
    (:pvnv (cffi:mem-ref p :nv))
    (:pvmg (error "Blessed or magical scalars not supported yet"))
    (:pvav (list-from-av p))
    (:pvhv (hash-from-hv p))))
(defun start-perl ()
  (unless *perl-interpreter*
    (setq *perl-interpreter* (make-interpreter)))
  (values))

(defun stop-perl ()
  (when *perl-interpreter*
    (destroy-interpreter *perl-interpreter*)
    (setq *perl-interpreter* nil))
  (values))
(defun call-perl (function return-type methodp &rest args)
  (need-perl)
  (perl-scope 
   (pushmark)
   (push-mortals-on-stack args)
   (get-stack-by-type
    return-type
    (funcall (if (stringp function) #'perl-call-function
               ;; either a scalar string or a code reference
               #'perl-call-scalar)
             function
             (calling-flags return-type methodp)))))
(defun use-perl-module (name &optional version)
  (eval-perl (format nil "use ~A ~@[~A~];" name version)))
